# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging
import re
from dataclasses import fields
from socket import gaierror

import more_itertools
from requests.exceptions import ConnectionError, ReadTimeout

import faebryk.library._F as F
from atopile.errors import UserInfraError
from faebryk.core.module import Module
from faebryk.core.parameter import And, Is, Parameter, ParameterOperatable
from faebryk.core.solver.solver import LOG_PICK_SOLVE, Solver
from faebryk.libs.exceptions import UserException, downgrade
from faebryk.libs.picker.api.api import ApiHTTPError, get_api_client
from faebryk.libs.picker.api.models import (
    BaseParams,
    CapacitorParams,
    Component,
    # DiodeParams,
    # InductorParams,
    LCSCParams,
    # LDOParams,
    # LEDParams,
    ManufacturerPartParams,
    # MOSFETParams,
    ResistorParams,
    # TVSParams,
)
from faebryk.libs.picker.lcsc import (
    LCSC_NoDataException,
    LCSC_PinmapException,
    attach,
    check_attachable,
    get_raw,
)
from faebryk.libs.picker.picker import (
    NotCompatibleException,
    PickError,
    does_not_require_picker_check,
)
from faebryk.libs.sets.sets import P_Set
from faebryk.libs.test.times import Times
from faebryk.libs.util import (
    Tree,
    cast_assert,
    groupby,
    not_none,
)

logger = logging.getLogger(__name__)
client = get_api_client()


type SIvalue = str

# TODO add way for user to specify quantity of PCBAs
qty: int = 1


class PickerUnboundedParameterError(Exception):
    pass


class PickerESeriesIntersectionError(Exception):
    pass


def _extract_numeric_id(lcsc_id: str) -> int:
    match = re.match(r"C(\d+)", lcsc_id)
    if match is None:
        raise ValueError(f"Invalid LCSC part number {lcsc_id}")
    return int(match[1])


TYPE_SPECIFIC_LOOKUP: dict[F.is_pickable_by_type.Type, type[BaseParams]] = {
    F.is_pickable_by_type.Type.Resistor: ResistorParams,
    F.is_pickable_by_type.Type.Capacitor: CapacitorParams,
    # F.is_pickable_by_type.Type.Inductor: InductorParams,
    # F.is_pickable_by_type.Type.TVS: TVSParams,
    # F.is_pickable_by_type.Type.LED: LEDParams,
    # F.is_pickable_by_type.Type.Diode: DiodeParams,
    # F.is_pickable_by_type.Type.LDO: LDOParams,
    # F.is_pickable_by_type.Type.MOSFET: MOSFETParams,
}


def _prepare_query(
    module: Module, solver: Solver
) -> BaseParams | LCSCParams | ManufacturerPartParams:
    assert module.has_trait(F.is_pickable)
    # Error can propagate through,
    # because we expect all pickable modules to be attachable
    check_attachable(module)

    if trait := module.try_get_trait(F.is_pickable_by_part_number):
        return ManufacturerPartParams(
            manufacturer_name=trait.get_manufacturer(),
            part_number=trait.get_partno(),
            quantity=qty,
        )

    elif trait := module.try_get_trait(F.is_pickable_by_supplier_id):
        if trait.get_supplier() == F.is_pickable_by_supplier_id.Supplier.LCSC:
            return LCSCParams(
                lcsc=_extract_numeric_id(trait.get_supplier_part_id()), quantity=qty
            )

    elif trait := module.try_get_trait(F.is_pickable_by_type):
        pick_type = trait.get_pick_type()
        params_t = TYPE_SPECIFIC_LOOKUP[pick_type]

        if pkg_t := module.try_get_trait(F.has_package):
            package = pkg_t.package.get_last_known_deduced_superset(solver)
        else:
            package = None

        generic_field_names = {f.name for f in fields(params_t)}
        _, known_params = more_itertools.partition(
            lambda p: p.get_name() in generic_field_names, module.get_parameters()
        )
        cmp_params = {
            p.get_name(): p.get_last_known_deduced_superset(solver)
            for p in known_params
        }

        if all(superset is None for superset in cmp_params.values()):
            logger.warning(f"Module `{module}` has no constrained parameters")

        return params_t(package=package, qty=qty, **cmp_params)  # type: ignore

    raise NotImplementedError(
        f"Unsupported pickable trait: {module.get_trait(F.is_pickable)}"
    )


def _process_candidates(module: Module, candidates: list[Component]) -> list[Component]:
    # Filter parts with weird pinmaps
    it = iter(candidates)
    filtered_candidates = []
    for c in it:
        try:
            attach(module, c.lcsc_display, check_only=True, get_model=False)
            filtered_candidates.append(c)
            # If we found one that's ok, just continue since likely enough
            filtered_candidates.extend(it)
            break
        except LCSC_NoDataException:
            if len(candidates) == 1:
                raise
        except LCSC_PinmapException:
            # if all filtered by pinmap something is fishy
            if not filtered_candidates and candidates[-1] is c:
                raise

    return filtered_candidates


def _find_modules(
    modules: Tree[Module], solver: Solver
) -> dict[Module, list[Component]]:
    timings = Times(name="find_modules")

    params = {m: _prepare_query(m, solver) for m in modules}
    timings.add("prepare queries")

    grouped = groupby(params.items(), lambda p: p[1])
    queries = list(grouped.keys())

    def _map_response[T](results: list[T]) -> dict[Module, T]:
        assert len(results) == len(queries)
        return {m: r for ms, r in zip(grouped.values(), results) for m, _ in ms}

    try:
        results = client.fetch_parts_multiple(queries)
        timings.add("fetch parts")
    except ConnectionError as e:
        cause = e.args[0]
        while not isinstance(cause, gaierror):
            cause = cause.__cause__
            if cause is None:
                break
        else:
            raise UserInfraError(
                f"Fetching component data failed: connection error: {cause.strerror}"
            ) from e

        raise UserInfraError("Fetching component data failed: connection error") from e
    except ReadTimeout as e:
        raise UserInfraError(
            "Fetching component data failed to complete in time. "
            "Please try again later."
        ) from e
    except ApiHTTPError as e:
        if e.response.status_code == 400:
            response = cast_assert(dict, e.response.json())
            if errors := response.get("detail", {}).get("errors", None):
                raise ExceptionGroup(
                    "Failed to fetch one or more parts",
                    [
                        PickError(f"{error['message']}\n{query.pretty_str()}", module)
                        for module, (query, error) in _map_response(
                            list(zip(queries, errors))
                        ).items()
                        if error is not None
                    ],
                ) from e
            else:
                raise
        raise e

    out = {m: _process_candidates(m, r) for m, r in _map_response(results).items()}
    timings.add("process candidates")
    return out


def _attach(module: Module, c: Component):
    """
    Calls LCSC attach and wraps errors into PickError
    """

    try:
        c.attach(module)
    except (
        ValueError,
        Component.ParseError,
        LCSC_NoDataException,
        LCSC_PinmapException,
    ) as e:
        if LOG_PICK_SOLVE:
            logger.warning(f"Failed to attach {c} to `{module}`: {e}")

        raise PickError(
            f"Failed to attach component {c} to module `{module}`: {e}",
            module,
        )


def _get_compatible_parameters(
    module: Module, c: "Component", solver: Solver
) -> dict[Parameter, ParameterOperatable.Literal]:
    """
    Check if the parameters of a component are compatible with the module
    """
    # Nothing to check
    if not module.has_trait(F.is_pickable_by_type):
        return {}

    # shortcut because solving slow
    try:
        get_raw(c.lcsc_display)
    except LCSC_NoDataException as e:
        raise NotCompatibleException(module, c) from e

    design_params = module.get_trait(F.is_pickable_by_type).get_parameters()
    component_params = c.attribute_literals

    if no_attr := component_params.keys() - design_params.keys():
        with downgrade(UserException):
            no_attr_str = "\n".join(f"- `{a}`" for a in no_attr)
            raise UserException(
                f"Module `{module}` is missing attributes:\n\n"
                f" {no_attr_str}\n\n"
                "This likely means you could use a more precise"
                " module/component in your design."
            )

    def _map_param(name: str, param: Parameter) -> tuple[Parameter, P_Set]:
        c_range = component_params.get(name)
        if c_range is None:
            c_range = param.domain.unbounded(param)
        return param, c_range

    param_mapping = [
        _map_param(name, param)
        for name, param in design_params.items()
        if not param.has_trait(does_not_require_picker_check)
    ]

    # check for any param that has few supersets whether the component's range
    # is compatible already instead of waiting for the solver
    for m_param, c_range in param_mapping:
        # TODO other loglevel
        # logger.warning(f"Checking obvious incompatibility for param {m_param}")
        known_superset = solver.inspect_get_known_supersets(m_param)
        if not known_superset.is_superset_of(c_range):
            if LOG_PICK_SOLVE:
                logger.warning(
                    f"Known superset {known_superset} is not a superset of {c_range}"
                    f" for part C{c.lcsc}"
                )
            raise NotCompatibleException(module, c, m_param, c_range)

    return {p: c_range for p, c_range in param_mapping}


def _check_candidates_compatible(
    module_candidates: list[tuple[Module, Component]],
    solver: Solver,
    allow_not_deducible: bool = False,
):
    """
    Check if combination of all candidates is compatible with each other
    Checks each candidate first for individual compatibility
    """

    if not module_candidates:
        return

    mappings = [_get_compatible_parameters(m, c, solver) for m, c in module_candidates]

    if LOG_PICK_SOLVE:
        logger.info(f"Solving for modules:" f" {[m for m, _ in module_candidates]}")

    predicates = (
        Is(m_param, c_range)
        for param_mapping in mappings
        for m_param, c_range in not_none(param_mapping).items()
    )

    solver.try_fulfill(And(*predicates), lock=False, allow_unknown=allow_not_deducible)


# public -------------------------------------------------------------------------------


def check_and_attach_candidates(
    candidates: list[tuple[Module, Component]],
    solver: Solver,
    allow_not_deducible: bool = False,
):
    """
    Check if given candidates are compatible with each other
    If so, attach them to the modules
    Raises:
        Contradiction
        NotCompatibleException
    """
    _check_candidates_compatible(candidates, solver, allow_not_deducible)

    for m, part in candidates:
        _attach(m, part)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                f"Attached {part.lcsc_display} ('{part.description}') to "
                f"'{m.get_full_name(types=False)}'"
            )


def get_candidates(
    modules: Tree[Module], solver: Solver
) -> dict[Module, list[Component]]:
    candidates = modules.copy()
    parts = {}
    empty = set()

    while candidates:
        # TODO deduplicate parts with same literals
        new_parts = _find_modules(modules, solver)

        parts.update({m: p for m, p in new_parts.items() if p})
        empty = {m for m, p in new_parts.items() if not p}
        for m in parts:
            if m in candidates:
                candidates.pop(m)
        if not empty:
            return parts
        for m in empty:
            subtree = candidates.pop(m)
            if not subtree:
                raise PickError(
                    f"No candidates found for `{m}`:\n{m.pretty_params(solver)}", m
                )
            candidates.update(subtree)

    # should fail earlier
    return {}


def attach_single_no_check(cmp: Module, part: Component, solver: Solver):
    """
    Attach a single component to a module
    Attention: Does not check compatibility before or after!
    """
    _attach(cmp, part)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


from enum import IntEnum, StrEnum

import faebryk.library._F as F
from faebryk.core.parameter import ParameterOperatable
from faebryk.libs.library import L
from faebryk.libs.units import P


class ButtonCell(F.Battery):
    class Material(StrEnum):
        Alkaline = "L"
        SilverOxide = "S"
        ZincAir = "P"
        Lithium = "C"
        Mercury = "M"
        NickelCadmium = "K"
        NickelMetalHydride = "H"

        @property
        def voltage(self) -> ParameterOperatable.NumberLike:
            return {
                self.Alkaline: 1.5 * P.V,
                self.SilverOxide: 1.55 * P.V,
                self.ZincAir: 1.65 * P.V,
                self.Lithium: 3.0 * P.V,
                self.Mercury: 1.35 * P.V,
                self.NickelCadmium: 1.2 * P.V,
                self.NickelMetalHydride: 1.2 * P.V,
            }[self]

    class Shape(StrEnum):
        Round = "R"

    class Size(IntEnum):
        L_41 = 41
        L_43 = 43
        L_44 = 44
        L_54 = 54
        N_1025 = 1025
        N_1216 = 1216
        N_1220 = 1220
        N_1225 = 1225
        N_1616 = 1616
        N_1620 = 1620
        N_1632 = 1632
        N_2016 = 2016
        N_2025 = 2025
        N_2032 = 2032
        N_2430 = 2430
        N_2450 = 2450

    material = L.p_field(
        domain=L.Domains.ENUM(Material),
    )
    shape = L.p_field(
        domain=L.Domains.ENUM(Shape),
    )
    size = L.p_field(
        domain=L.Domains.ENUM(Size),
    )

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.B
    )

    # TODO merge voltage with material voltage



# pylint: disable=logging-fstring-interpolation

"""
This CLI command provides the `ato install` command to:
- install dependencies
- download JLCPCB footprints
"""

import logging
import shutil
from pathlib import Path
from typing import Annotated, Optional
from urllib.parse import quote, urlparse

import questionary
import requests
import ruamel.yaml
import typer
from git import GitCommandError, InvalidGitRepositoryError, NoSuchPathError, Repo

import faebryk.libs.exceptions
from atopile import errors, version
from atopile.config import Dependency, ProjectConfig, Source, config
from faebryk.libs.util import robustly_rm_dir

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
yaml = ruamel.yaml.YAML()


def install(
    to_install: Annotated[str | None, typer.Argument()] = None,
    jlcpcb: Annotated[
        bool, typer.Option("--jlcpcb", "-j", help="JLCPCB component ID", hidden=True)
    ] = False,
    vendor: Annotated[
        bool,
        typer.Option(
            "--vendor",
            help="Copy the contents of this dependency into the repo",
        ),
    ] = False,
    local: Annotated[
        Path | None, typer.Option("--local", "-l", help="Install from local path")
    ] = None,
    upgrade: Annotated[
        bool, typer.Option("--upgrade", "-u", help="Upgrade dependencies")
    ] = False,
    path: Annotated[Path | None, typer.Argument()] = None,
):
    """
    Install atopile packages or components from jlcpcb.com/parts
    """
    if jlcpcb:
        raise errors.UserBadParameterError(
            "--jlcpcb flag has been replaced by `ato create component`"
        )

    config.apply_options(None)

    do_install(to_install, vendor, upgrade, path, local)


def do_install(
    to_install: str | None,
    vendor: bool,
    upgrade: bool,
    path: Path | None,
    local: Path | None,
):
    """
    Actually do the installation of the dependencies.
    This is split in two so that it can be called from `install` and `create`
    """

    if path is not None:
        config.project_dir = path

    if to_install is None:
        logger.info(f"Installing all dependencies in {config.project.paths.root}")
    else:
        logger.info(f"Installing {to_install} in {config.project.paths.root}")

    if local:
        # eg. "ato install --local /path/to/local/module local_module"
        install_single_local(local, name=to_install)
    elif to_install:
        # eg. "ato install some-atopile-module"
        install_single_dependency(to_install, vendor, upgrade)
    else:
        # eg. "ato install"
        install_project_dependencies(upgrade)

    logger.info("[green]Done![/] :call_me_hand:", extra={"markup": True})


def get_package_repo_from_registry(module_name: str) -> str:
    """
    Get the git repo for a package from the ato registry.
    """
    try:
        encoded_name = quote(module_name)
        response = requests.get(
            f"{config.project.services.packages.url}/v0/package/{encoded_name}",
            timeout=10,
        )
    except requests.exceptions.ReadTimeout as ex:
        raise errors.UserInfraError(
            f"Request to registry timed out for package '{module_name}'"
        ) from ex

    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as ex:
        try:
            _ = response.json()["detail"]
            if response.status_code == 404:
                raise errors.UserException(
                    f"Could not find package '{module_name}' in registry.",
                    markdown=False,
                ) from None
        except (KeyError, requests.exceptions.JSONDecodeError):
            pass

        raise errors.UserException(
            f"Error getting data for package '{module_name}': \n{ex}",
            markdown=False,
        )

    return_data = response.json()

    try:
        return_url = return_data["data"]["repo_url"]
    except KeyError as ex:
        raise errors.UserException(
            f"No repo_url found for package '{module_name}'"
        ) from ex
    return return_url


def add_dependency(config_data, new_data):
    config_data["dependencies"] = [
        dep.model_dump()
        # add_dependencies is the field validator that loads the dependencies
        # from the config file. It ensures the format of the ato.yaml
        for dep in ProjectConfig.add_dependencies(
            config_data.get("dependencies"),
        )  # type: ignore  add_dependencies is a classmethod
    ]

    for i, dep in enumerate(config_data["dependencies"]):
        if dep["name"] == new_data["name"]:
            config_data["dependencies"][i] = new_data
            break
    else:
        config_data["dependencies"] = config_data["dependencies"] + [new_data]

    return config_data


def install_single_local(path: Path, name: str | None = None):
    name = name or path.name
    dependency = Dependency(
        name=name,
        source=Source(local=path),
        path=(config.project.paths.modules / name).relative_to(
            config.project.paths.root
        ),
        project_config=ProjectConfig.from_path(path),
    )
    install_local_dependency(dependency)
    config.update_project_config(add_dependency, dependency.model_dump())


def install_single_dependency(to_install: str, vendor: bool, upgrade: bool):
    dependency = Dependency.from_str(to_install)
    name = _name_and_clone_url_helper(dependency.name)[0]
    if vendor:
        dependency.link_broken = True
        abs_path = config.project.paths.src / name
        dependency.path = abs_path.relative_to(config.project.paths.root)
    else:
        dependency.link_broken = False
        abs_path = config.project.paths.modules / name
        dependency.path = abs_path.relative_to(config.project.paths.root)

    try:
        installed_version = install_dependency(dependency, upgrade, abs_path)
    except GitCommandError as ex:
        if "already exists and is not an empty directory" in ex.stderr:
            # FIXME: shouldn't `--upgrade` do this already?
            raise errors.UserException(
                f"Directory {abs_path} already exists and is not empty. "
                "Please move or remove it before installing this new content."
            ) from ex
        raise
    # If the link's broken, remove the .git directory so git treats it as copy-pasted code # noqa: E501  # pre-existing
    if dependency.link_broken:
        try:
            robustly_rm_dir(abs_path / ".git")
        except (PermissionError, OSError, FileNotFoundError) as ex:
            with faebryk.libs.exceptions.downgrade(errors.UserException):
                raise errors.UserException(
                    f"Failed to remove .git directory: {repr(ex)}"
                ) from ex

    if dependency.version_spec is None and installed_version:
        # If the user didn't specify a version, we'll
        # use the one we just installed as a basis
        dependency.version_spec = f"@{installed_version}"

    config.update_project_config(add_dependency, dependency.model_dump())


def install_project_dependencies(upgrade: bool):
    for _ctx, dependency in faebryk.libs.exceptions.iter_through_errors(
        config.project.dependencies or []
    ):
        with _ctx():
            if dependency.source and dependency.source.local:
                install_local_dependency(dependency)
                continue

            if not dependency.link_broken:
                # FIXME: these dependency objects are a little too entangled
                name = _name_and_clone_url_helper(dependency.name)[0]
                abs_path = config.project.paths.modules / name
                dependency.path = abs_path.relative_to(config.project.paths.root)

                try:
                    install_dependency(dependency, upgrade, abs_path)
                except GitCommandError as ex:
                    if "already exists and is not an empty directory" in ex.stderr:
                        # FIXME: shouldn't `--upgrade` do this already?
                        raise errors.UserException(
                            f"Directory {abs_path} already exists and is not empty. "
                            "Please move or remove it before installing this new content."  # noqa: E501  # pre-existing
                        ) from ex
                    raise


def install_local_dependency(dependency: Dependency):
    if not dependency.source or not dependency.source.local:
        raise errors.UserFileNotFoundError(
            "Local dependency must have a source with local path"
        )
    src = dependency.source.local
    dst = dependency.path or config.project.paths.modules / dependency.name
    if not src.exists():
        raise errors.UserException(f"Local dependency path {src} does not exist")
    shutil.copytree(src, dst, dirs_exist_ok=True)


def install_dependency(
    dependency: Dependency, upgrade: bool, abs_path: Path
) -> Optional[str]:
    """
    Install a dependency of the name "module_name"
    """
    # Ensure the modules path exists
    abs_path.parent.mkdir(parents=True, exist_ok=True)

    # Figure out what we're trying to install here
    module_spec = dependency.version_spec or "*"
    module_name, clone_url = _name_and_clone_url_helper(dependency.name)

    try:
        # This will raise an exception if the directory does not exist
        repo = Repo(abs_path)
    except (InvalidGitRepositoryError, NoSuchPathError):
        # Directory does not contain a valid repo, clone into it
        logger.info(f"Installing dependency `{module_name}`")
        repo = Repo.clone_from(clone_url, abs_path)
        tracking = repo.active_branch.tracking_branch()
        if tracking:
            tracking.checkout()
        else:
            logger.warning(
                f"No tracking branch found for {module_name}, using current branch"
            )
    else:
        # In this case the directory exists and contains a valid repo
        if upgrade:
            logger.info(f"Fetching latest changes for {module_name}")
            repo.remotes.origin.fetch()
        else:
            logger.info(
                f"{module_name} already exists. If you wish to upgrade, use --upgrade"
            )
            # here we're done because we don't want to play with peoples' deps under them # noqa: E501  # pre-existing
            return

    # Figure out what version of this thing we need
    semver_to_tag = {}
    installed_semver = None
    for tag in repo.tags:
        try:
            semver_to_tag[version.parse(tag.name)] = tag
        except errors.UserException:
            logger.debug(f"Tag {tag.name} is not a valid semver tag. Skipping.")

    if "@" in module_spec:
        # If there's an @ in the version, we're gonna check that thing out
        best_checkout = module_spec.strip(" @")
    elif semver_to_tag:
        # Otherwise we're gonna find the best tag meeting the semver spec
        valid_versions = [v for v in semver_to_tag if version.match(module_spec, v)]
        if not valid_versions:
            raise errors.UserException(
                f"No versions of {module_name} match spec {module_spec}.\n"
                f"Available versions: {', '.join(map(str, semver_to_tag))}"
            )
        installed_semver = max(valid_versions)
        best_checkout = semver_to_tag[installed_semver]
    else:
        logger.warning(
            "No semver tags found for this module. Using latest default branch :hot_pepper:.",  # noqa: E501  # pre-existing
            extra={"markup": True},
        )
        return None

    # If the repo is dirty, throw an error
    if repo.is_dirty():
        raise errors.UserException(
            f"Module {module_name} has uncommitted changes. Aborting."
        )

    # Checkout the best thing we've found
    ref_before_checkout = repo.head.commit

    # If the repo best_checkout is a branch, we need to checkout the origin/branch
    if best_checkout in repo.heads:
        best_checkout = f"origin/{best_checkout}"

    repo.git.checkout(best_checkout)

    if repo.head.commit == ref_before_checkout:
        logger.info(
            f"Already on the best option ([cyan bold]{best_checkout}[/]) for {module_name}",  # noqa: E501  # pre-existing
            extra={"markup": True},
        )
    else:
        logger.info(
            f"Using :sparkles: [cyan bold]{best_checkout}[/] :sparkles: of {module_name}",  # noqa: E501  # pre-existing
            extra={"markup": True},
        )

    return repo.head.commit.hexsha


def check_missing_deps() -> bool:
    for dependency in config.project.dependencies or []:
        if dependency.path:
            dep_path = config.project.paths.root / dependency.path
        else:
            # FIXME: this should exist based on defaults in the config
            dep_path = config.project.paths.modules / dependency.name

        if not dep_path.exists():
            return True

    return False


def check_missing_deps_or_offer_to_install():
    if check_missing_deps():
        logger.warning(
            "It appears some dependencies are missing."
            " Run `ato install` to install them.",
            extra={"markdown": True},
        )

        if (
            config.interactive
            and questionary.confirm("Install missing dependencies now?").unsafe_ask()
        ):
            # Install project dependencies, without upgrading
            install_project_dependencies(False)


def _name_and_clone_url_helper(name: str) -> tuple[str, str]:
    """Return the name of the package and the URL to clone it from."""
    parsed_url = urlparse(name)
    if not parsed_url.scheme:
        return name, get_package_repo_from_registry(name)
    else:
        splitted = parsed_url.path.split("/")
        return splitted[-1] or splitted[-2], name



from pathlib import Path
from typing import Sequence

from antlr4 import CommonTokenStream, ParserRuleContext, Token
from rich.console import Console, ConsoleOptions, ConsoleRenderable
from rich.highlighter import ReprHighlighter
from rich.markdown import Markdown
from rich.syntax import Syntax
from rich.text import Text

from atopile.parse_utils import (
    PygmentsLexerReconstructor,
    get_src_info_from_token,
)
from faebryk.core.node import NodeException
from faebryk.libs.exceptions import UserException as _BaseBaseUserException


def _render_tokens(
    token_stream: CommonTokenStream, start_token: Token, stop_token: Token
) -> list[ConsoleRenderable]:
    lexer = PygmentsLexerReconstructor.from_tokens(
        token_stream, start_token, stop_token, 1, 1
    )
    src_path, src_line, src_col = get_src_info_from_token(start_token)

    # Make the path relative to the current working directory, if possible
    try:
        src_path = Path(src_path).relative_to(Path.cwd())
    except ValueError:
        pass
    source_info = str(src_path)
    if src_line := src_line:
        source_info += f":{src_line}"
    if src_col := src_col:
        source_info += f":{src_col}"

    highlight_lines = set(range(start_token.line, stop_token.line + 1))

    return [
        Text("Source: ", style="bold") + Text(source_info, style="magenta"),
        Syntax(
            lexer.get_code(),
            lexer,
            line_numbers=True,
            start_line=lexer.start_line,
            indent_guides=True,
            highlight_lines=highlight_lines,
        ),
    ]


class _BaseUserException(_BaseBaseUserException):
    """
    This exception is thrown when there's an error in the syntax of the language
    """

    def __init__(
        self,
        msg: str,
        *args,
        token_stream: CommonTokenStream | None = None,
        origin_start: Token | None = None,
        origin_stop: Token | None = None,
        traceback: Sequence[ParserRuleContext | None] | None = None,
        markdown: bool = True,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)

        self.token_stream = token_stream
        self.origin_start = origin_start
        self.origin_stop = origin_stop
        self.traceback = traceback
        self.markdown = markdown

        if self.token_stream is None and (
            self.origin_start is not None or self.origin_stop is not None
        ):
            raise ValueError(
                "token_stream is required if origin_start or origin_stop is provided"
            )

    def attach_origin_from_ctx(self, ctx: ParserRuleContext) -> None:
        self.origin_start = ctx.start
        self.origin_stop = ctx.stop
        self.token_stream = ctx.parser.getInputStream()  # type: ignore

    @classmethod
    def from_ctx[T: _BaseUserException](
        cls: type[T],
        origin: ParserRuleContext | None,
        message: str,
        *args,
        traceback: Sequence[ParserRuleContext | None] | None = None,
        **kwargs,
    ) -> T:
        """Create an error from a context."""

        instance = cls(message, *args, traceback=traceback, **kwargs)

        if origin is not None:
            instance.attach_origin_from_ctx(origin)

        return instance

    @classmethod
    def from_tokens[T: _BaseUserException](
        cls: type[T],
        token_stream: CommonTokenStream,
        origin_start: Token,
        origin_stop: Token | None,
        message: str,
        *args,
        traceback: Sequence[ParserRuleContext | None] | None = None,
        **kwargs,
    ) -> T:
        """Create an error from a context."""

        instance = cls(
            message,
            *args,
            token_stream=token_stream,
            origin_start=origin_start,
            origin_stop=origin_stop,
            traceback=traceback,
            **kwargs,
        )
        return instance

    def get_frozen(self) -> tuple:
        if self.origin_start and self.origin_stop:
            return (
                super().get_frozen()
                + get_src_info_from_token(self.origin_start)
                + get_src_info_from_token(self.origin_stop)
            )
        return super().get_frozen()

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> list[ConsoleRenderable]:
        def _markdown(message: str) -> ConsoleRenderable:
            if console.is_terminal:
                return Markdown(message)
            else:
                return Text(message)

        renderables: list[ConsoleRenderable] = []

        if self.title:
            renderables += [Text(self.title, style="bold")]

        renderables += [
            _markdown(self.message)
            if self.markdown
            else ReprHighlighter()(Text(self.message))
        ]

        for ctx in self.traceback or []:
            if ctx is not None:
                assert self.token_stream is not None
                renderables += _render_tokens(self.token_stream, ctx.start, ctx.stop)

        if self.origin_start:
            assert self.token_stream is not None

            if self.origin_stop is None:
                origin_stop = self.origin_start
            else:
                origin_stop = self.origin_stop

            renderables += [Text("Code causing the error: ", style="bold")]
            renderables += _render_tokens(
                self.token_stream, self.origin_start, origin_stop
            )

        return renderables


class UserFatalException(_BaseUserException):
    """
    Something in the user's code meant we weren't able to continue.
    Don't display a traceback on these because we'll have already printed one.
    """


class UserException(_BaseUserException):
    """
    This exception is thrown when there's an error in ato code
    """


class UserSyntaxError(UserException):
    """
    Raised when there's an error in the syntax of the language
    """


class UserKeyError(UserException, KeyError):
    """
    Raised if a name isn't found in the current scope.
    """


class UserTypeError(UserException):
    """
    Raised if something is the wrong type.
    """


class UserValueError(UserException):
    """
    Raised if something is the correct type but an invalid value.
    """


class UserImportNotFoundError(UserException):
    """
    Raised if something has a conflicting name in the same scope.
    """


class UserAmbiguousReferenceError(UserException):
    """
    Raised if something has a conflicting name in the same scope.
    """


class UserFileNotFoundError(UserException, FileNotFoundError):
    """
    Raised if a file couldn't be found.
    """


class UserUnknownUnitError(UserException):
    """
    Raised if a unit couldn't be interpreted.
    """


class UserIncompatibleUnitError(UserException):
    """
    Raised if a unit couldn't be interpreted.
    """


class UserInfraError(UserException):
    """
    Raised when there's an issue contacting atopile
    infrastructure needed for an operation.
    """

    title = "Infrastructure Error"


class UserNotImplementedError(UserException):
    """
    Raised when a feature is not yet implemented.
    """


class UserBadParameterError(UserException):
    """
    Raised when a bad CLI param is given
    """


class UserPythonLoadError(UserException):
    """
    Raised when a Python module couldn't be loaded.
    """


class UserPythonModuleError(UserException):
    """
    Raised when a user-provided Python module is faulty.
    """


class UserPythonConstructionError(UserPythonModuleError):
    """
    Raised when a Python module couldn't be constructed.
    """


class UserAssertionError(UserException):
    """
    Raised when an assertion fails.
    """


class UserContradictionException(UserException):
    """
    Raised when user-provided constraints contradict.
    """


class UserPickError(UserException):
    """
    Raised when there's an error in the picker.
    """


class UserActionWithoutEffectError(UserException):
    """
    Raised when an action is performed but has no effect.
    """


class UserAlreadyExistsError(UserException):
    """
    Raised when something already exists.
    """


class UserNodeException(UserException):
    """
    Raised when there's an error with a node operation.
    """

    @classmethod
    def from_node_exception(
        cls,
        node_ex: NodeException,
        origin: ParserRuleContext,
        traceback: Sequence[ParserRuleContext | None] | None,
        *args,
        **kwargs,
    ) -> "UserNodeException":
        return cls.from_ctx(
            origin,
            node_ex.message,
            *args,
            traceback=traceback,
            **kwargs,
        )


class UserNoProjectException(UserException):
    """
    Raised when the project directory is not found.
    """

    def __init__(
        self,
        msg: str = (
            "Could not find the project directory, are you within an ato project?"
        ),
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT


import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.solver.defaultsolver import DefaultSolver
from faebryk.exporters.bom.jlcpcb import _get_bomline
from faebryk.libs.app.designators import attach_random_designators, load_designators
from faebryk.libs.app.pcb import create_footprint_library
from faebryk.libs.iso_metric_screw_thread import Iso262_MetricScrewThreadSizes
from faebryk.libs.library import L
from faebryk.libs.picker.picker import pick_part_recursively
from faebryk.libs.units import P


def _build(app: Module):
    load_designators(app.get_graph(), attach=True)
    solver = DefaultSolver()
    pick_part_recursively(app, solver)
    F.has_package.standardize_footprints(app, solver)
    create_footprint_library(app, no_fp_lib=True)
    attach_random_designators(app.get_graph())


def test_bom_mounting_hole():
    mh = F.Mounting_Hole(
        diameter=Iso262_MetricScrewThreadSizes.M10,
        pad_type=F.Mounting_Hole.PadType.Pad,
    )

    _build(mh)

    bomline = _get_bomline(mh)
    assert bomline is None


def test_bom_picker_pick():
    r = F.Resistor()
    r.resistance.constrain_subset(L.Range.from_center_rel(10 * P.kohm, 0.01))

    _build(r)

    bomline = _get_bomline(r)
    assert bomline is not None


def test_bom_explicit_pick():
    m = Module()
    m.add(F.can_attach_to_footprint_symmetrically())
    m.add(F.has_explicit_part.by_supplier("C25804", pinmap={}))
    _build(m)

    bomline = _get_bomline(m)
    assert bomline is not None
    assert bomline.LCSC_Partnumber == "C25804"


def test_bom_kicad_footprint_no_lcsc():
    m = Module()
    m.add(F.can_attach_to_footprint_symmetrically())
    fp = F.KicadFootprint(pin_names=["1", "2"])
    fp.add(
        F.KicadFootprint.has_kicad_identifier(
            "Connector_PinHeader_2.54mm:PinHeader_1x02_P2.54mm_Vertical"
        )
    )
    m.get_trait(F.can_attach_to_footprint).attach(fp)
    _build(m)

    bomline = _get_bomline(m)
    assert bomline is None


def test_bom_kicad_footprint_lcsc_verbose():
    m = Module()
    m.add(F.can_attach_to_footprint_symmetrically())
    fp = F.KicadFootprint(pin_names=["1", "2"])
    fp.add(
        F.KicadFootprint.has_kicad_identifier(
            "Connector_PinHeader_2.54mm:PinHeader_1x02_P2.54mm_Vertical"
        )
    )
    m.get_trait(F.can_attach_to_footprint).attach(fp)
    m.add(F.has_explicit_part.by_supplier("C18166021", pinmap={}))

    _build(m)

    bomline = _get_bomline(m)
    assert bomline is not None
    assert bomline.LCSC_Partnumber == "C18166021"
    assert m.get_trait(F.has_footprint).get_footprint() is fp


def test_bom_kicad_footprint_lcsc_compact():
    m = Module()
    m.add(F.can_attach_to_footprint_symmetrically())
    m.add(
        F.has_explicit_part.by_supplier(
            "C18166021",
            pinmap={},
            override_footprint=(
                F.KicadFootprint(pin_names=["1", "2"]),
                "Connector_PinHeader_2.54mm:PinHeader_1x02_P2.54mm_Vertical",
            ),
        )
    )

    _build(m)

    bomline = _get_bomline(m)
    assert bomline is not None
    assert bomline.LCSC_Partnumber == "C18166021"
    assert (
        m.get_trait(F.has_footprint)
        .get_footprint()
        .get_trait(F.has_kicad_footprint)
        .get_kicad_footprint_name()
        == "PinHeader_1x02_P2.54mm_Vertical"
    )



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.libs.library import L
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class UART_RS485(Module):
    power: F.ElectricPower
    uart: F.UART_Base
    rs485: F.RS485HalfDuplex
    read_enable: F.ElectricLogic
    write_enable: F.ElectricLogic

    max_data_rate = L.p_field(units=P.baud)
    gpio_voltage = L.p_field(units=P.V)

    def __preinit__(self):
        self.max_data_rate.alias_is(self.uart.baud)
        self.power.voltage.constrain_subset(L.Range(3.3 * P.V, 5.0 * P.V))
        # FIXME
        # self.power.decoupled.decouple()

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )



import logging
from pathlib import Path

import pcbnew  # type: ignore

from .common import (
    footprints_by_addr,
    get_footprint_addr,
    get_layout_map,
    groups_by_name,
    log_exceptions,
)

log = logging.getLogger(__name__)


class ReloadGroup(pcbnew.ActionPlugin):
    def defaults(self):
        self.name = "Reload Group"
        self.category = "Reload Group Layout Atopile"
        self.description = (
            "Layout components on PCB in same spatial"
            " relationships as components on schematic"
        )
        self.show_toolbar_button = True
        self.icon_file_name = str(Path(__file__).parent / "reload.png")
        self.dark_icon_file_name = self.icon_file_name

    @log_exceptions()
    def Run(self):
        board: pcbnew.BOARD = pcbnew.GetBoard()
        board_path = board.GetFileName()

        groups = groups_by_name(board)
        log.debug(f"Named groups in boards: {groups}")

        # We can ignore the uuid map here because in the context of the parent module,
        # there will always be addresses with the v0.3 compiler
        footprints = footprints_by_addr(board, {})

        for group_name, group_data in get_layout_map(board_path).items():
            log.debug(f"Updating group {group_name}")

            # FIXME: we rely on the fact that this dict is topologically sorted
            # from the layout.py code to ensure that nested groups are created
            # prior to their parent groups

            # If the group doesn't yet exist in the layout
            # create it and add it to the board
            if group_name in groups:
                g = groups[group_name]
                log.debug(f"Group {group_name} already exists. Using it.")
            else:
                g = pcbnew.PCB_GROUP(board)
                g.SetName(group_name)
                board.Add(g)
                groups[group_name] = g
                log.debug(f"Group {group_name} created and added to board")

            # Make sure all the footprints in the group are up to date
            footprints_in_group = {
                addr
                for fp in g.GetItems()
                if isinstance(fp, pcbnew.FOOTPRINT)
                and (addr := get_footprint_addr(fp, {}))
                and addr in footprints
            }
            expected_footprints = set(group_data["group_components"])

            for fp_addr in footprints_in_group - expected_footprints:
                g.RemoveItem(footprints[fp_addr])
                log.debug(f"Removed footprint {fp_addr}")

            # Add all items to the group
            # Start with the footprints
            for fp_addr in expected_footprints - footprints_in_group:
                if fp_addr in footprints:
                    g.AddItem(footprints[fp_addr])

            # FIXME: nested groups are not yet supported
            if group_data["nested_groups"]:
                raise NotImplementedError("Nested groups are not yet supported")


with log_exceptions():
    ReloadGroup().register()



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F  # noqa: F401
from faebryk.libs.library import L  # noqa: F401
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P  # noqa: F401

logger = logging.getLogger(__name__)


class TECH_PUBLIC_USBLC6_2SC6(F.USB2_0_ESD_Protection):
    """
    USB 2.0 ESD protection

    6A 12V 150W 6V Unidirectional 5V SOT-23-6 ESD and Surge Protection (TVS/ESD) ROHS
    """

    # ----------------------------------------
    #     modules, interfaces, parameters
    # ----------------------------------------

    # ----------------------------------------
    #                 traits
    # ----------------------------------------
    lcsc_id = L.f_field(F.has_descriptive_properties_defined)({"LCSC": "C2827654"})
    descriptive_properties = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "TECH PUBLIC",
            DescriptiveProperties.partno: "USBLC6-2SC6",
        }
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://www.lcsc.com/datasheet/lcsc_datasheet_2108132230_TECH-PUBLIC-USBLC6-2SC6_C2827654.pdf"
    )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.usb[0].usb_if.buspower.lv: ["GND"],
                self.usb[0].usb_if.d.n.line: ["IO1"],
                self.usb[0].usb_if.d.p.line: ["IO2"],
                self.usb[0].usb_if.buspower.hv: ["VBUS"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )

    def __preinit__(self):
        # ------------------------------------
        #           connections
        # ------------------------------------

        # ------------------------------------
        #          parametrization
        # ------------------------------------
        pass



from pathlib import Path

import pytest

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.exporters.schematic.kicad.transformer import SchTransformer
from faebryk.libs.exceptions import UserException
from faebryk.libs.kicad.fileformats_sch import C_kicad_sch_file
from faebryk.libs.test.fileformats import FPLIBFILE, SCHFILE


@pytest.fixture
def fp_lib_path_path():
    return FPLIBFILE


@pytest.fixture
def sch_file():
    return C_kicad_sch_file.loads(SCHFILE)


@pytest.fixture
def transformer(sch_file: C_kicad_sch_file):
    app = Module()
    return SchTransformer(sch_file.kicad_sch, app.get_graph(), app)


def test_wire_transformer(transformer: SchTransformer):
    start_wire_count = len(transformer.sch.wires)

    transformer.insert_wire(
        [
            (0, 0),
            (1, 0),
            (1, 1),
        ]
    )

    # 2 because we have 3 waypoints
    assert len(transformer.sch.wires) == start_wire_count + 2
    assert [(pt.x, pt.y) for pt in transformer.sch.wires[-2].pts.xys] == [
        (0, 0),
        (1, 0),
    ]


def test_index_symbol_files(transformer: SchTransformer, fp_lib_path_path: Path):
    assert len(transformer._symbol_files_index) == 0
    transformer.index_symbol_files(fp_lib_path_path, load_globals=False)
    assert len(transformer._symbol_files_index) == 1


@pytest.fixture
def full_transformer(transformer: SchTransformer, fp_lib_path_path: Path):
    transformer.index_symbol_files(fp_lib_path_path, load_globals=False)
    return transformer


def test_get_symbol_file(full_transformer: SchTransformer):
    with pytest.raises(UserException):
        full_transformer.get_symbol_file("notta-lib")

    sym_flie = full_transformer.get_symbol_file("test")
    assert (
        sym_flie.kicad_symbol_lib.symbols["AudioJack-CUI-SJ-3523-SMT"].name
        == "AudioJack-CUI-SJ-3523-SMT"
    )


def test_insert_symbol(full_transformer: SchTransformer):
    start_symbol_count = len(full_transformer.sch.symbols)

    # mimicing typically design/user-space
    audio_jack = full_transformer.app.add(Module())
    pin_s = audio_jack.add(F.Electrical())
    pin_t = audio_jack.add(F.Electrical())
    pin_r = audio_jack.add(F.Electrical())
    audio_jack.add(F.has_overriden_name_defined("U1"))

    # mimicing typically lcsc code
    sym = F.Symbol.with_component(
        audio_jack,
        {
            "S": pin_s,
            "T": pin_t,
            "R": pin_r,
        },
    )
    audio_jack.add(F.Symbol.has_symbol(sym))
    sym.add(F.Symbol.has_kicad_symbol("test:AudioJack-CUI-SJ-3523-SMT"))

    full_transformer.insert_symbol(audio_jack)

    assert len(full_transformer.sch.symbols) == start_symbol_count + 1
    assert full_transformer.sch.symbols[-1].propertys["Reference"].value == "U1"



import logging
import shutil
from collections.abc import Iterable
from datetime import datetime
from pathlib import Path
from types import ModuleType, TracebackType

import pathvalidate
from rich._null_file import NullFile
from rich.columns import Columns
from rich.console import Console, ConsoleRenderable, RenderableType
from rich.live import Live
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.padding import Padding
from rich.spinner import Spinner
from rich.table import Table
from rich.text import Text
from rich.traceback import Traceback

import atopile
import faebryk
import faebryk.libs
import faebryk.libs.logging
from atopile.errors import UserPythonModuleError, _BaseBaseUserException
from faebryk.libs.logging import FLOG_FMT
from faebryk.libs.util import ConfigFlag

from . import console

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

COLOR_LOGS = ConfigFlag("COLOR_LOGS", default=False)
NOW = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
_DEFAULT_FORMATTER = logging.Formatter("%(message)s", datefmt="[%X]")
_SHOW_LOG_FILE_PATH_THRESHOLD = 120


# displayed during LoggingStage
ALERT = logging.INFO + 5
logging.addLevelName(ALERT, "ALERT")


class LogHandler(RichHandler):
    """
    A logging handler that renders output with Rich.

    Suppresses frames from tracebacks conditionally depending on the exception type.
    """

    def __init__(
        self,
        *args,
        console: Console,
        rich_tracebacks: bool = True,
        show_path: bool = False,
        tracebacks_suppress: Iterable[str] | None = ["typer"],
        tracebacks_suppress_map: dict[type[BaseException], Iterable[ModuleType]]
        | None = {UserPythonModuleError: [atopile, faebryk]},
        tracebacks_unwrap: list[type[BaseException]] | None = [UserPythonModuleError],
        hide_traceback_types: tuple[type[BaseException], ...] = (
            _BaseBaseUserException,
        ),
        always_show_traceback_types: tuple[type[BaseException], ...] = (
            UserPythonModuleError,
        ),
        traceback_level: int = logging.ERROR,
        force_terminal: bool = False,
        **kwargs,
    ):
        super().__init__(
            *args,
            console=console,
            rich_tracebacks=rich_tracebacks,
            show_path=show_path,
            **kwargs,
        )
        self.tracebacks_suppress = tracebacks_suppress or []
        self.tracebacks_suppress_map = tracebacks_suppress_map or {}
        self.tracebacks_unwrap = tracebacks_unwrap or []
        self.hide_traceback_types = hide_traceback_types
        self.always_show_traceback_types = always_show_traceback_types
        self.traceback_level = traceback_level
        self._logged_exceptions = set()
        self._is_terminal = force_terminal or console.is_terminal

        self.addFilter(
            lambda record: record.name.startswith("atopile")
            or record.name.startswith("faebryk")
        )

    def _get_suppress(
        self, exc_type: type[BaseException] | None
    ) -> Iterable[str | ModuleType]:
        """
        Get extended list of modules to suppress from tracebacks.
        """
        suppress: set[str | ModuleType] = set(self.tracebacks_suppress)
        if exc_type is not None:
            for _type in self.tracebacks_suppress_map:
                if issubclass(exc_type, _type) or isinstance(exc_type, _type):
                    suppress.update(self.tracebacks_suppress_map[_type])
        return list(suppress)

    def _unwrap_chained_exceptions(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> tuple[type[BaseException] | None, BaseException | None, TracebackType | None]:
        def unwrap(exc: BaseException, exc_type: type[BaseException]) -> BaseException:
            """
            Unwraps an exception chain until we reach an exception
            that is not an instance of _type.
            """
            while isinstance(exc, exc_type) and exc.__cause__ is not None:
                exc = exc.__cause__
            return exc

        if exc_type is not None and exc_value is not None:
            while any(issubclass(exc_type, _type) for _type in self.tracebacks_unwrap):
                exc_value = unwrap(exc_value, exc_type)
                exc_type = type(exc_value)
                exc_traceback = exc_value.__traceback__

        return exc_type, exc_value, exc_traceback

    def _get_traceback(self, record: logging.LogRecord) -> Traceback | None:
        if not record.exc_info:
            return None

        exc_type, exc_value, exc_traceback = record.exc_info

        suppress = self._get_suppress(exc_type)

        hide_traceback = (
            isinstance(exc_value, self.hide_traceback_types)
            and not isinstance(exc_value, self.always_show_traceback_types)
        ) or record.levelno < self.traceback_level

        if isinstance(exc_value, UserPythonModuleError):
            exc_type, exc_value, exc_traceback = self._unwrap_chained_exceptions(
                exc_type, exc_value, exc_traceback
            )

        if hide_traceback or exc_type is None or exc_value is None:
            return None

        return Traceback.from_exception(
            exc_type,
            exc_value,
            exc_traceback,
            width=self.tracebacks_width,
            extra_lines=self.tracebacks_extra_lines,
            theme=self.tracebacks_theme,
            word_wrap=self.tracebacks_word_wrap,
            show_locals=self.tracebacks_show_locals,
            locals_max_length=self.locals_max_length,
            locals_max_string=self.locals_max_string,
            suppress=suppress,
        )

    def _render_message(
        self, record: logging.LogRecord, message: str
    ) -> "ConsoleRenderable":
        """Render message text in to Text.

        Args:
            record (LogRecord): logging Record.
            message (str): String containing log message.

        Returns:
            ConsoleRenderable: Renderable to display log message.
        """
        if not self._is_terminal:
            return Text(message)

        use_markdown = getattr(record, "markdown", False)
        use_markup = getattr(record, "markup", self.markup)
        highlighter = getattr(record, "highlighter", self.highlighter)

        if use_markdown:
            message_text = Markdown(message)
        else:
            message_text = Text.from_markup(message) if use_markup else Text(message)

            if highlighter:
                message_text = highlighter(message_text)

            if self.keywords is None:
                self.keywords = self.KEYWORDS

            if self.keywords:
                message_text.highlight_words(self.keywords, "logging.keyword")

        return message_text

    def render_message(
        self, record: logging.LogRecord, message: str
    ) -> ConsoleRenderable:
        if record.exc_info is not None and isinstance(
            (exc := record.exc_info[1]), ConsoleRenderable
        ):
            # UserExceptions are already renderables
            return exc

        return self._render_message(record, message)

    def _get_hashable(self, record: logging.LogRecord) -> tuple | None:
        if exc_info := getattr(record, "exc_info", None):
            _, exc_value, _ = exc_info
            if exc_value and isinstance(exc_value, _BaseBaseUserException):
                return exc_value.get_frozen()
        return None

    def _prepare_emit(self, record: logging.LogRecord) -> ConsoleRenderable:
        traceback = self._get_traceback(record)

        if self.formatter:
            record.message = record.getMessage()
            formatter = self.formatter
            if hasattr(formatter, "usesTime") and formatter.usesTime():
                record.asctime = formatter.formatTime(record, formatter.datefmt)
            message = formatter.formatMessage(record)
        else:
            message = record.getMessage()

        message_renderable = self.render_message(record, message)

        log_renderable = self.render(
            record=record, traceback=traceback, message_renderable=message_renderable
        )

        return log_renderable

    def emit(self, record: logging.LogRecord) -> None:
        """Invoked by logging."""
        hashable = self._get_hashable(record)
        if hashable and hashable in self._logged_exceptions:
            return None

        log_renderable = self._prepare_emit(record)

        if isinstance(self.console.file, NullFile):
            # Handles pythonw, where stdout/stderr are null, and we return NullFile
            # instance from Console.file. In this case, we still want to make a log
            # record even though we won't be writing anything to a file.
            self.handleError(record)
        else:
            try:
                self.console.print(log_renderable)
            except Exception:
                self.handleError(record)
            finally:
                if hashable:
                    self._logged_exceptions.add(hashable)


class LiveLogHandler(LogHandler):
    def __init__(self, status: "LoggingStage", *args, **kwargs):
        super().__init__(*args, console=status._console, **kwargs)
        self.status = status

    def emit(self, record: logging.LogRecord) -> None:
        hashable = self._get_hashable(record)
        if hashable and hashable in self._logged_exceptions:
            return

        try:
            if record.levelno >= logging.ERROR:
                self.status._error_count += 1
            elif record.levelno >= logging.WARNING:
                self.status._warning_count += 1

            self.status._live.update(self.status._render_status(), refresh=True)

            if record.levelno == ALERT:
                self.status.alert(record.getMessage())

        except Exception:
            self.handleError(record)
        finally:
            if hashable:
                self._logged_exceptions.add(hashable)


class LoggingStage:
    _INDICATOR_SUCCESS = "[green]✓[/green]"
    _INDICATOR_FAILURE = "[red]✗[/red]"
    _INDICATOR_WARNING = "[yellow]⚠[/yellow]"

    _LOG_LEVELS = {
        logging.DEBUG: "debug",
        logging.INFO: "info",
        logging.WARNING: "warning",
        logging.ERROR: "error",
    }

    def __init__(self, name: str, description: str, indent: int = 20):
        self.name = name
        self.description = description
        self.indent = indent
        self._console = console.error_console
        self._spinner = Spinner("dots")
        self._warning_count = 0
        self._error_count = 0
        self._info_log_path = None
        self._log_handler = None
        self._file_handlers = []
        self._original_handlers = {}
        self._live = Live(
            self._render_status(),
            console=self._console,
            transient=True,
            auto_refresh=True,
            refresh_per_second=10,
        )
        self._sanitized_name = pathvalidate.sanitize_filename(self.name)
        self._result = None

    def alert(self, message: str) -> None:
        message = f"[bold][yellow]![/yellow] {message}[/bold]"

        self._console.print(
            Padding(Text.from_markup(message), pad=(0, 0, 0, self.indent)),
            highlight=True,
        )

    def _render_status(self, indicator: str | None = None) -> RenderableType:
        problems = []
        if self._error_count > 0:
            plural_e = "s" if self._error_count > 1 else ""
            problems.append(f"[red]{self._error_count} error{plural_e}[/red]")

        if self._warning_count > 0:
            plural_w = "s" if self._warning_count > 1 else ""
            problems.append(f"[yellow]{self._warning_count} warning{plural_w}[/yellow]")

        problems_text = f" ({', '.join(problems)})" if problems else ""
        text = Text.from_markup(f"{self.description}{problems_text}")

        spinner_with_text = Padding(
            Columns(
                [
                    Text.from_markup(indicator)
                    if indicator is not None
                    else self._spinner,
                    text,
                ],
                padding=(0, 1),
            ),
            pad=(0, 0, 0, self.indent),  # (top, right, bottom, left)
        )

        if self._info_log_path and self._console.width >= _SHOW_LOG_FILE_PATH_THRESHOLD:
            table = Table.grid(padding=0, expand=True)
            table.add_column(max_width=self.indent + 32)
            table.add_column(justify="right", overflow="ellipsis")
            table.add_row(spinner_with_text, f"[dim]{self._info_log_path}[/dim]")
            return table
        else:
            return spinner_with_text

    def __enter__(self) -> Console:
        self._setup_logging()
        self._live.start()
        return self._console

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._restore_logging()

        if exc_type is not None or self._error_count > 0:
            indicator = self._INDICATOR_FAILURE
        elif self._warning_count > 0:
            indicator = self._INDICATOR_WARNING
        else:
            indicator = self._INDICATOR_SUCCESS

        self._live.stop()
        self._console.print(self._render_status(indicator))

    def _create_log_dir(self) -> Path:
        from atopile.config import config

        base_log_dir = Path(config.project.paths.logs) / NOW

        try:
            build_cfg = config.build
            log_dir = base_log_dir / pathvalidate.sanitize_filename(build_cfg.name)
        except RuntimeError:
            log_dir = base_log_dir

        log_dir.mkdir(parents=True, exist_ok=True)

        latest_link = Path(config.project.paths.logs) / "latest"
        if latest_link.exists():
            if latest_link.is_symlink():
                latest_link.unlink()
            else:
                shutil.rmtree(latest_link)
        latest_link.symlink_to(base_log_dir, target_is_directory=True)

        return log_dir

    def _setup_logging(self) -> None:
        root_logger = logging.getLogger()

        self._original_level = root_logger.level
        self._original_handlers = {"root": root_logger.handlers.copy()}

        # log all messages to at least one handler
        root_logger.setLevel(logging.DEBUG)

        self._log_handler = LiveLogHandler(self)
        self._log_handler.setFormatter(_DEFAULT_FORMATTER)
        self._log_handler.setLevel(self._original_level)

        log_dir = self._create_log_dir()

        for handler in root_logger.handlers.copy():
            root_logger.removeHandler(handler)

        root_logger.addHandler(self._log_handler)

        self._file_handlers = []
        self._file_handles = {}

        for level, level_name in self._LOG_LEVELS.items():
            log_file = log_dir / f"{self._sanitized_name}.{level_name}.log"

            self._file_handles[level_name] = log_file.open("w")
            file_console = Console(
                file=self._file_handles[level_name],
                width=500,
                force_terminal=COLOR_LOGS.get(),
            )
            file_handler = LogHandler(
                console=file_console,
                force_terminal=COLOR_LOGS.get(),
                markup=False,
            )
            file_handler.setFormatter(_DEFAULT_FORMATTER)
            file_handler.setLevel(level)
            self._file_handlers.append(file_handler)
            root_logger.addHandler(file_handler)

            if level_name == "info":
                self._info_log_path = log_file.relative_to(Path.cwd())

    def _restore_logging(self) -> None:
        if not self._log_handler and not self._file_handlers:
            return

        root_logger = logging.getLogger()
        root_logger.setLevel(self._original_level)
        if self._log_handler in root_logger.handlers:
            root_logger.removeHandler(self._log_handler)

        for file_handler in self._file_handlers:
            if file_handler in root_logger.handlers:
                root_logger.removeHandler(file_handler)
                file_handler.close()

        for file_handle in self._file_handles.values():
            file_handle.close()

        for handler in root_logger.handlers.copy():
            root_logger.removeHandler(handler)

        for handler in self._original_handlers.get("root", []):
            root_logger.addHandler(handler)

        self._original_handlers = {}
        self._original_level = logging.INFO
        self._log_handler = None
        self._file_handlers = []


handler = LogHandler(console=console.error_console)
handler.setFormatter(_DEFAULT_FORMATTER)

if FLOG_FMT:
    faebryk.libs.logging.setup_basic_logging()
else:
    logging.basicConfig(level=logging.INFO, handlers=[handler])

logger = logging.getLogger(__name__)



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

# This file is auto-generated by nanobind.
# Do not edit this file directly; edit the corresponding
# C++ file instead.
import enum
from collections.abc import Callable, Sequence, Set
from typing import overload

class Counter:
    @property
    def in_cnt(self) -> int: ...
    @property
    def weak_in_cnt(self) -> int: ...
    @property
    def out_weaker(self) -> int: ...
    @property
    def out_stronger(self) -> int: ...
    @property
    def out_cnt(self) -> int: ...
    @property
    def time_spent_s(self) -> float: ...
    @property
    def hide(self) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def multi(self) -> bool: ...
    @property
    def total_counter(self) -> bool: ...

class Edge:
    def __repr__(self) -> str: ...
    @property
    def to(self) -> GraphInterface: ...

class Graph:
    def __init__(self) -> None: ...
    def get_edges(self, arg: GraphInterface, /) -> dict[GraphInterface, Link]: ...
    @property
    def edges(self) -> list[tuple[GraphInterface, GraphInterface, Link]]: ...
    def get_gifs(self) -> set[GraphInterface]: ...
    def invalidate(self) -> None: ...
    @property
    def node_count(self) -> int: ...
    @property
    def edge_count(self) -> int: ...
    def node_projection(self) -> set[Node]: ...
    def nodes_by_names(self, arg: Set[str], /) -> list[tuple[Node, str]]: ...
    def bfs_visit(
        self,
        filter: Callable[[Sequence[GraphInterface], Link], bool],
        start: Sequence[GraphInterface],
    ) -> set[GraphInterface]: ...
    def __repr__(self) -> str: ...

class GraphInterface:
    def __init__(self) -> None: ...
    def __repr__(self) -> str: ...
    def get_graph(self) -> Graph: ...
    @property
    def G(self) -> Graph: ...
    def get_gif_edges(self) -> set[GraphInterface]: ...
    @property
    def edges(self) -> dict[GraphInterface, Link]: ...
    @property
    def node(self) -> Node: ...
    @node.setter
    def node(self, arg: Node, /) -> None: ...
    def is_connected_to(self, arg: GraphInterface, /) -> Link | None: ...
    @property
    def name(self) -> str: ...
    @name.setter
    def name(self, arg: str, /) -> None: ...
    def get_connected_nodes(self, types: Sequence[type]) -> set[Node]: ...
    @overload
    def connect(self, other: GraphInterface) -> None: ...
    @overload
    def connect(self, others: Sequence[GraphInterface]) -> None: ...
    @overload
    def connect(self, other: GraphInterface, link: Link) -> None: ...
    @overload
    def connect(self, others: Sequence[GraphInterface], link: Link) -> None: ...

class GraphInterfaceHierarchical(GraphInterface):
    def __init__(self, is_parent: bool) -> None: ...
    def get_parent(self) -> tuple[Node, str] | None: ...
    def get_children(self) -> list[Node]: ...
    @property
    def is_parent(self) -> bool: ...
    def disconnect_parent(self) -> None: ...

class GraphInterfaceModuleConnection(GraphInterface):
    def __init__(self) -> None: ...

class GraphInterfaceModuleSibling(GraphInterfaceHierarchical):
    def __init__(self, is_parent: bool) -> None: ...

class GraphInterfaceReference(GraphInterface):
    def __init__(self) -> None: ...
    def get_referenced_gif(self) -> GraphInterfaceSelf: ...
    def get_reference(self) -> Node: ...

class GraphInterfaceReferenceUnboundError(Exception):
    pass

class GraphInterfaceSelf(GraphInterface):
    def __init__(self) -> None: ...

class Link:
    def __eq__(self, arg: Link, /) -> bool: ...
    def is_cloneable(self) -> bool: ...

class LinkDirect(Link):
    def __init__(self) -> None: ...

class LinkDirectConditional(LinkDirect):
    def __init__(
        self,
        filter: Callable[[Path], LinkDirectConditionalFilterResult],
        needs_only_first_in_path: bool = False,
    ) -> None: ...

class LinkDirectConditionalFilterResult(enum.Enum):
    FILTER_PASS = 0

    FILTER_FAIL_RECOVERABLE = 1

    FILTER_FAIL_UNRECOVERABLE = 2

class LinkDirectDerived(LinkDirectConditional):
    def __init__(self, arg: Path, /) -> None: ...

class LinkExists(Exception):
    pass

class LinkFilteredException(Exception):
    pass

class LinkNamedParent(LinkParent):
    def __init__(self, arg: str, /) -> None: ...

class LinkParent(Link):
    def __init__(self) -> None: ...

class LinkPointer(Link):
    def __init__(self) -> None: ...

class LinkSibling(LinkPointer):
    def __init__(self) -> None: ...

class Node:
    def __init__(self) -> None: ...
    @staticmethod
    def transfer_ownership(arg: Node, /) -> Node: ...
    def get_graph(self) -> Graph: ...
    @property
    def self_gif(self) -> GraphInterfaceSelf: ...
    @property
    def children(self) -> GraphInterfaceHierarchical: ...
    @property
    def parent(self) -> GraphInterfaceHierarchical: ...
    def get_children(
        self,
        direct_only: bool,
        types: Sequence[type] | None = None,
        include_root: bool = False,
        f_filter: Callable[[Node], bool] | None = None,
        sort: bool = True,
    ) -> list[Node]: ...
    def get_parent(self) -> tuple[Node, str] | None: ...
    def get_parent_force(self) -> tuple[Node, str]: ...
    def get_name(self, accept_no_parent: bool = False) -> str: ...
    def get_hierarchy(self) -> list[tuple[Node, str]]: ...
    def get_full_name(self, types: bool = False) -> str: ...
    @overload
    def isinstance(self, type: type) -> bool: ...
    @overload
    def isinstance(self, types: Sequence[type]) -> bool: ...
    def bfs_node(self, filter: Callable[[Path], bool]) -> set[Node]: ...
    @property
    def no_include_parents_in_full_name(self) -> bool: ...
    @no_include_parents_in_full_name.setter
    def no_include_parents_in_full_name(self, arg: bool, /) -> None: ...
    def __repr__(self) -> str: ...

class NodeException(Exception):
    pass

class NodeNoParent(Exception):
    pass

class Path:
    def __repr__(self) -> str: ...
    def __len__(self) -> int: ...
    def contains(self, arg: GraphInterface, /) -> bool: ...
    def last(self) -> GraphInterface: ...
    def first(self) -> GraphInterface: ...
    def get_link(self, arg: Edge, /) -> Link: ...
    def iterate_edges(self, arg: Callable[[Edge], bool], /) -> None: ...
    def __getitem__(self, arg: int, /) -> GraphInterface: ...

def add(i: int, j: int = 1) -> int:
    """A function that adds two numbers"""

def call_python_function(func: Callable[[], int]) -> int: ...
def find_paths(src: Node, dst: Sequence[Node]) -> tuple[list[Path], list[Counter]]: ...
def print_obj(obj: object) -> None: ...
def set_indiv_measure(value: bool) -> None: ...
def set_leak_warnings(value: bool) -> None: ...
def set_max_paths(arg0: int, arg1: int, arg2: int, /) -> None: ...



# This file is part of the faebryk project
# SPDX-License-Identifier: MIT

import logging

import faebryk.library._F as F
from faebryk.core.module import Module
from faebryk.core.moduleinterface import ModuleInterface
from faebryk.libs.library import L
from faebryk.libs.picker.picker import DescriptiveProperties
from faebryk.libs.units import P

logger = logging.getLogger(__name__)


class RP2040(Module):
    """
    Raspberry Pi RP2040 MCU
    Note: Don't forget to use the pinmux!
    """

    class PIO(F.ElectricLogic):
        pass

    class ADC(F.ElectricSignal):
        pass

    class PWM(ModuleInterface):
        A: F.ElectricLogic
        B: F.ElectricLogic

        @L.rt_field
        def single_electric_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

    class CoreRegulator(Module):
        power_in: F.ElectricPower
        power_out: F.ElectricPower

        def __preinit__(self):
            F.ElectricLogic.connect_all_module_references(self, gnd_only=True)

            # TODO get tolerance
            self.power_out.voltage.constrain_subset(
                L.Range.from_center_rel(1.1 * P.V, 0.05)
            )
            self.power_in.voltage.constrain_subset(L.Range(1.8 * P.V, 3.3 * P.V))

        @L.rt_field
        def bridge(self):
            return F.can_bridge_defined(self.power_in, self.power_out)

    class SPI(F.SPI):
        cs: F.ElectricLogic

    class UART(ModuleInterface):
        base_uart: F.UART_Base
        rts: F.ElectricLogic
        cts: F.ElectricLogic

        @L.rt_field
        def single_electric_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

    class USBPowerControl(ModuleInterface):
        ovcur_det: F.ElectricLogic
        vbus_det: F.ElectricLogic
        vbus_en: F.ElectricLogic

        @L.rt_field
        def single_reference(self):
            return F.has_single_electric_reference_defined(
                F.ElectricLogic.connect_all_module_references(self)
            )

    # power
    power_io: F.ElectricPower
    power_adc: F.ElectricPower
    power_core: F.ElectricPower
    power_usb_phy: F.ElectricPower
    core_regulator: CoreRegulator

    io = L.list_field(30, F.Electrical)
    io_soft = L.list_field(6, F.Electrical)

    # IO
    qspi = L.f_field(F.MultiSPI)(data_lane_count=4)
    swd: F.SWD
    xtal_if: F.XtalIF

    run: F.ElectricLogic
    usb: F.USB2_0_IF.Data
    factory_test_enable: F.Electrical
    usb_power_control: USBPowerControl

    # peripherals
    spi = L.list_field(2, SPI)
    pwm = L.list_field(8, PWM)
    i2c = L.list_field(2, F.I2C)
    uart = L.list_field(2, UART)
    pio = L.list_field(2, PIO)
    adc = L.list_field(4, ADC)
    clock_in = L.list_field(2, F.ElectricLogic)
    clock_out = L.list_field(4, F.ElectricLogic)
    gpio = L.list_field(30 + 6, F.ElectricLogic)

    def __preinit__(self):
        # TODO get tolerance
        self.power_adc.voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.05)
        )
        self.power_usb_phy.voltage.constrain_subset(
            L.Range.from_center_rel(3.3 * P.V, 0.05)
        )
        self.power_core.voltage.constrain_subset(
            L.Range.from_center_rel(1.1 * P.V, 0.05)
        )
        self.power_io.voltage.constrain_subset(L.Range(1.8 * P.V, 3.3 * P.V))

        F.ElectricLogic.connect_all_module_references(self, gnd_only=True)
        F.ElectricLogic.connect_all_node_references(
            [self.power_io]
            + self.gpio
            + self.spi
            + self.pwm
            + self.uart
            + self.i2c
            + self.pio
            + self.clock_in
            + self.clock_out
            + [self.usb_power_control]  # TODO is this the right ref?
            + [self.run]
            + [self.swd]
            + [self.qspi]
        )
        self.power_io.lv.connect(self.xtal_if.gnd)

        # QSPI pins reusable for soft gpio
        self.qspi.data[3].line.connect(self.io_soft[0])
        self.qspi.clock.line.connect(self.io_soft[1])
        self.qspi.data[0].line.connect(self.io_soft[2])
        self.qspi.data[2].line.connect(self.io_soft[3])
        self.qspi.data[1].line.connect(self.io_soft[4])
        self.qspi.chip_select.line.connect(self.io_soft[5])

        # ADC pins shared with GPIO
        self.adc[0].line.connect(self.io[26])
        self.adc[1].line.connect(self.io[27])
        self.adc[2].line.connect(self.io[28])
        self.adc[3].line.connect(self.io[29])
        F.ElectricLogic.connect_all_node_references(self.adc + [self.power_adc])

    @L.rt_field
    def pinmux(self):
        return F.RP2040Pinmux(self)

    @L.rt_field
    def decoupled(self):
        return F.can_be_decoupled_rails(self.power_io, self.power_core)

    designator_prefix = L.f_field(F.has_designator_prefix)(
        F.has_designator_prefix.Prefix.U
    )
    datasheet = L.f_field(F.has_datasheet_defined)(
        "https://datasheets.raspberrypi.com/rp2040/rp2040-datasheet.pdf"
    )

    mfr = L.f_field(F.has_descriptive_properties_defined)(
        {
            DescriptiveProperties.manufacturer: "Raspberry Pi",
            DescriptiveProperties.partno: "RP2040",
        }
    )

    @L.rt_field
    def attach_to_footprint(self):
        return F.can_attach_to_footprint_via_pinmap(
            {
                "1": self.power_io.hv,
                "2": self.io[0],
                "3": self.io[1],
                "4": self.io[2],
                "5": self.io[3],
                "6": self.io[4],
                "7": self.io[5],
                "8": self.io[6],
                "9": self.io[7],
                "10": self.power_io.hv,
                "11": self.io[8],
                "12": self.io[9],
                "13": self.io[10],
                "14": self.io[11],
                "15": self.io[12],
                "16": self.io[13],
                "17": self.io[14],
                "18": self.io[15],
                "19": self.factory_test_enable,
                "20": self.xtal_if.xin,
                "21": self.xtal_if.xout,
                "22": self.power_io.hv,
                "23": self.power_core.hv,
                "24": self.swd.clk.line,
                "25": self.swd.dio.line,
                "26": self.run.line,
                "27": self.io[16],
                "28": self.io[17],
                "29": self.io[18],
                "30": self.io[19],
                "31": self.io[20],
                "32": self.io[21],
                "33": self.power_io.hv,
                "34": self.io[22],
                "35": self.io[23],
                "36": self.io[24],
                "37": self.io[25],
                "38": self.io[26],
                "39": self.io[27],
                "40": self.io[28],
                "41": self.io[29],
                "42": self.power_io.hv,
                "43": self.power_adc.hv,
                "44": self.core_regulator.power_in.hv,
                "45": self.core_regulator.power_out.hv,
                "46": self.usb.n.line,
                "47": self.usb.p.line,
                "48": self.power_usb_phy.hv,
                "49": self.power_io.hv,
                "50": self.power_core.hv,
                "51": self.qspi.data[3].line,
                "52": self.qspi.clock.line,
                "53": self.qspi.data[0].line,
                "54": self.qspi.data[2].line,
                "55": self.qspi.data[1].line,
                "56": self.qspi.chip_select.line,
                "57": self.power_io.lv,  # center pad
            }
        )

    @L.rt_field
    def pin_association_heuristic(self):
        return F.has_pin_association_heuristic_lookup_table(
            mapping={
                self.power_adc.hv: ["ADC_AVDD"],
                self.power_core.hv: ["DVDD"],
                self.power_io.lv: ["GND"],
                self.io[0]: ["GPIO0"],
                self.io[1]: ["GPIO1"],
                self.io[10]: ["GPIO10"],
                self.io[11]: ["GPIO11"],
                self.io[12]: ["GPIO12"],
                self.io[13]: ["GPIO13"],
                self.io[14]: ["GPIO14"],
                self.io[15]: ["GPIO15"],
                self.io[16]: ["GPIO16"],
                self.io[17]: ["GPIO17"],
                self.io[18]: ["GPIO18"],
                self.io[19]: ["GPIO19"],
                self.io[2]: ["GPIO2"],
                self.io[20]: ["GPIO20"],
                self.io[21]: ["GPIO21"],
                self.io[22]: ["GPIO22"],
                self.io[23]: ["GPIO23"],
                self.io[24]: ["GPIO24"],
                self.io[25]: ["GPIO25"],
                self.io[26]: ["GPIO26_ADC0"],
                self.io[27]: ["GPIO27_ADC1"],
                self.io[28]: ["GPIO28_ADC2"],
                self.io[29]: ["GPIO29_ADC3"],
                self.io[3]: ["GPIO3"],
                self.io[4]: ["GPIO4"],
                self.io[5]: ["GPIO5"],
                self.io[6]: ["GPIO6"],
                self.io[7]: ["GPIO7"],
                self.io[8]: ["GPIO8"],
                self.io[9]: ["GPIO9"],
                self.power_io.hv: ["IOVDD"],
                self.qspi.clock.line: ["QSPI_SCLK"],
                self.qspi.data[0].line: ["QSPI_SD0"],
                self.qspi.data[1].line: ["QSPI_SD1"],
                self.qspi.data[2].line: ["QSPI_SD2"],
                self.qspi.data[3].line: ["QSPI_SD3"],
                self.qspi.chip_select.line: ["QSPI_SS", "QSPI_SS_N"],
                self.run.line: ["RUN"],
                self.swd.clk.line: ["SWCLK"],
                self.swd.dio.line: ["SWD"],
                self.factory_test_enable: ["TESTEN"],
                self.usb.n.line: ["USB_DM"],
                self.usb.p.line: ["USB_DP"],
                self.power_usb_phy.hv: ["USB_VDD"],
                self.core_regulator.power_in.hv: ["VREG_IN"],
                self.core_regulator.power_out.hv: ["VREG_VOUT"],
                self.xtal_if.xin: ["XIN"],
                self.xtal_if.xout: ["XOUT"],
            },
            accept_prefix=False,
            case_sensitive=False,
        )
